package Clone;
use strict;

our $VERSION = "1.3";

############################################################################
#
# fun: maakt nieuw Clone object, zie documentatie onderin
#
############################################################################

sub new
{
  my $pkg  = shift;
  my $self = {};
  bless $self,$pkg;
  return($self);
}

############################################################################
#
#  in: text
# out: data
# fun: zet string geserialiseerde data om in feitelijke datastructuur
# aan: $data=Clone::fromtext($text)
#
############################################################################

sub fromtext
{
  my $text       = &checktext(shift);
  return(undef) unless ($text);

  my ($top_element_key,$type2token) = eval($text);

  foreach my $type (keys %$type2token)
  {
    if ($type =~ /^V:/)
    {
      $type2token->{$type}=&unhexbytes($type2token->{$type});
    }
  }

  foreach my $type (keys %$type2token)
  {
    if ($type =~ /^(S:|R:)/)
    {
      $type2token->{$type}=\$type2token->{$type2token->{$type}};
    }
  }

  foreach my $type (keys %$type2token)
  {
    if ($type =~ /^A:/)
    {
      foreach my $elt (@{$type2token->{$type}})
      {
        $elt=$type2token->{$elt};
      }
    }
    elsif ($type =~ /^H:/)
    {
      foreach my $key (keys %{$type2token->{$type}})
      {
        my $value=$type2token->{$type}->{$key};
	delete $type2token->{$type}->{$key};
	$type2token->{$type}->{$type2token->{$key}}=$type2token->{$value};
      }
    }
    elsif ($type=~/^P:([^:]+):/)
    {
      my $package=$1;
      foreach my $key (keys %{$type2token->{$type}})
      {
        my $value=$type2token->{$type}->{$key};
	delete $type2token->{$type}->{$key};
        $type2token->{$type}->{$type2token->{$key}}=$type2token->{$value};
      }
      bless $type2token->{$type},$package;
    }
  }
  return($type2token->{$top_element_key}->{'Clone'});
}

############################################################################
#
#  in: data
# out: text
# fun: zet feitelijke datastructuur om in geserialiseerde data:
#      oorspronkelijke datastructuur wordt vernietigd
# aan: $text=Clone::convert2text($data)
#
############################################################################

sub convert2text
{
  my $token      = {'Clone' => shift};
  my $token2type = {};
  my $type2token = {};
  my $typenr     = {};

  &linear($token,$token2type,$type2token,$typenr);

  $token='('.&outnumber($token,$token2type,$type2token,$typenr).",\n".&text($type2token).")\n";

  return($token)
}

############################################################################
#
#  in: data
# out: text
# fun: zet feitelijke datastructuur om in geserialiseerde data
# aan: $text=Clone::copy2text($data)
#
############################################################################

sub copy2text
{
  my $token      = {'Clone' => &clone(shift)};
  my $token2type = {};
  my $type2token = {};
  my $typenr     = {};

  &linear($token,$token2type,$type2token,$typenr);

  return('('.&outnumber($token,$token2type,$type2token,$typenr).",\n".&text($type2token).")\n");
}

sub linear
{
  my $token      = shift;
  my $token2type = shift;
  my $type2token = shift;
  my $typenr     = shift;

  my $type       = ref $token;

  if ($type =~ /^(CODE|GLOB|LVALUE|IO::Handle)$/)
  {
    warn "type: $type is illegal, and cannot be copied from text";
    return();
  }

  unless ($type)
  {
    unless (defined($token))
    {
      $type = $token = 'UNDEF';
    }
    else
    {
      $type = 'VALUE';
    }
  }

  unless (exists($token2type->{$token}))
  {
    if ($type =~ /^(ARRAY|HASH|REF|SCALAR|VALUE|UNDEF)$/)
    {
      $type=substr($type,0,1);
    }
    else
    {
      $type= 'P:'.$type;
    }

    $typenr->{$type}++;
    $type.=':'.$typenr->{$type};

    $type2token->{$type}=$token;
    $token2type->{$token}=$type;

    if ($type =~ /^A:/)
    {
      foreach my $elt (@$token)
      {
        &linear($elt,$token2type,$type2token,$typenr);
      }
    }
    elsif ($type =~ /^H:/)
    {
      foreach my $elt (values %$token)
      {
        &linear($elt,$token2type,$type2token,$typenr);
      }
    }
    elsif ($type =~ /^(V:|U:)/)
    {
    }
    elsif ($type =~ /^(S:|R:)/)
    {
      &linear($$token,$token2type,$type2token,$typenr);
    }
    elsif ($type=~/^P:([^:]+):/)
    {
      foreach my $elt (values %$token)
      {
        &linear($elt,$token2type,$type2token,$typenr);
      }
    }
  }
}

sub outnumber
{
  my $token      = shift;
  my $token2type = shift;
  my $type2token = shift;
  my $typenr     = shift;

  foreach my $type (keys %$type2token)
  {
    if ($type =~ /^A/)
    {
      foreach my $elt (@{$type2token->{$type}})
      {
        $elt=$token2type->{$elt} if defined($elt);
      }
    }
    elsif ($type =~ /^H/)
    {

      foreach my $key (keys %{$type2token->{$type}})
      {
        unless (exists($token2type->{$key}))
        {
          $typenr->{'V'}++;
          $token2type->{$key}='V:'.$typenr->{'V'};
          $type2token->{'V:'.$typenr->{'V'}}=$key;
        }
        my $value=$type2token->{$type}->{$key};

	delete $type2token->{$type}->{$key};

        no warnings 'uninitialized';
	$type2token->{$type}->{$token2type->{$key}}=$token2type->{$value};
      }

    }
    elsif ($type =~ /^V/)
    {
    }
    elsif ($type =~ /^(S|R)/)
    {
      $type2token->{$type}=$token2type->{${$type2token->{$type}}};
    }
    elsif ($type=~/^P:([^:]+):/)
    {
      foreach my $key (keys %{$type2token->{$type}})
      {
        unless (exists($token2type->{$key}))
        {
          $typenr->{'V'}++;
          $token2type->{$key}='V:'.$typenr->{'V'};
          $type2token->{'V:'.$typenr->{'V'}}=$key;
        }
        my $value=$type2token->{$type}->{$key};

	delete $type2token->{$type}->{$key};
	$type2token->{$type}->{$token2type->{$key}}=$token2type->{$value};
      }
    }
  }

  my $type = ref($token);
  $type = 'VALUE' unless $type;

  if ($type =~ /^(ARRAY|HASH|REF|SCALAR|VALUE)$/)
  {
    $type=substr($type,0,1);
  }
  else
  {
    $type= 'P:'.$type;
  }

  return("'$type:1'");
}

sub text
{
  my $type2token = shift;
  my $text = '{';

  foreach my $type (keys %$type2token)
  {
    if ($type =~ /^A/)
    {
      $text.= "'$type'=>[";

      foreach my $elt (@{$type2token->{$type}})
      {

        if (defined($elt))
        {
	  $text .= "'$elt',";
        }
        else
        {
          $text .= "'',";
        }
      }
      $text =~ s/,$//;
      $text.="],";
    }
    elsif ($type =~ /^H/)
    {
      $text.= "'$type'=>{";

      no warnings 'uninitialized';
      foreach my $key (keys %{$type2token->{$type}})
      {
	$text .= "'$key'=>'$type2token->{$type}->{$key}',";
      }
      $text =~ s/,$//;
      $text.='},';
    }
    elsif ($type =~ /^V/)
    {
      $text.= "'$type'=>'".&hexbytes($type2token->{$type})."',";
    }
    elsif ($type =~ /^(S|R)/)
    {
      $text.= "'$type'=>'$type2token->{$type}',";
    }
    elsif ($type=~/^P:([^:]+):/)
    {
      if ($type2token->{$type} =~ /HASH/)
      {
        $text.="'$type'=>{";

        foreach my $key (keys %{$type2token->{$type}})
        {
	  $text .= "'$key'=>'$type2token->{$type}->{$key}',";
        }

        $text =~ s/,$//;
        $text.='},';
      }
      else
      {
        $text.= "'$type'=>[";

        foreach my $elt (@{$type2token->{$type}})
        {

          if (defined($elt))
          {
	    $text .= "'$elt',";
          }
          else
          {
            $text .= "'',";
          }
        }
        $text =~ s/,$//;
        $text.="],";
      }
    }
  }

  $text =~ s/,$/}/;
  return($text);
}

############################################################################
#
#  in: data
# out: data
# fun: produceert copy van datastructuur
# aan: $cloneobject->clone()
#
############################################################################

sub clone
{
  my $self      = shift;
  my $mem       = {};
  my $refs      = {};
  my $keys      = {};
  my $intersect = {};
  my $thorelli  = {};

  &refkeys($self,$keys,$refs);

  foreach my $key (keys %$refs)
  {
    $intersect->{$key}=1 if (exists($keys->{$key}));
  }

  my $copy = &u_copy($self,$mem);

  &thorelli($copy,$intersect,$mem,$thorelli);

  return($copy);
}

############################################################################
#
#  in: data
# out: data
# fun: produceert copy van datastructuur
# aan: $cloneobject->copy($data)
#
############################################################################

sub copy
{
  my $self      = shift;
  my $thing     = shift;
  my $mem       = {};
  my $refs      = {};
  my $keys      = {};
  my $intersect = {};

  &refkeys($thing,$keys,$refs);

  foreach my $key (keys %$refs)
  {
    $intersect->{$key}=1 if (exists($keys->{$key}));
  }

  my $copy = &u_copy($thing,$mem);
  &thorelli($copy,$intersect,$mem);

  return($copy);
}

sub refkeys
{
  my $thing     = shift;
  my $keys      = shift;
  my $refs      = shift;

  my $type   = ref $thing;


  if ($type  && $type !~ /SCALAR|CODE/)
  {
    no warnings 'uninitialized';
    $refs->{$thing}=1;

    if ($type =~ /REF/)
    {
      &refkeys($$thing,$keys,$refs) unless $refs->{$$thing};
    }
    elsif ($type =~ /ARRAY/)
    {
      foreach my $elt (@$thing)
      {
        &refkeys($elt,$keys,$refs) unless $refs->{$elt};
      }
    }
    else
    {
      foreach my $key (keys %$thing)
      {
        $keys->{$key}=1;
        &refkeys($thing->{$key},$keys,$refs) unless $refs->{$thing->{$key}};
      }
    }
  }
}

sub u_copy
{
  my $object = {};
  my $elt    = shift;
  my $mem    = shift;

  my $type   = ref $elt;

  unless ($type)
  {
    return($elt);
  }
  else
  {
    if ($type =~ /ARRAY/)
    {
      $mem->{$elt}=[];
      @{$mem->{$elt}}=&ar_copy($mem,@$elt);
    }
    elsif ($type =~ /HASH/)
    {
      $mem->{$elt}={};
      %{$mem->{$elt}}=&ha_copy($mem,$elt);
    }
    elsif ($type =~ /SCALAR|REF/)
    {
      $mem->{$elt}='';
      $mem->{$elt}=&sr_copy($mem,$$elt);
    }
    elsif ($type =~ /CODE/)
    {
      return($elt);
    }
    else
    {
      $mem->{$elt}={};
      %{$mem->{$elt}} = &ha_copy($mem,$elt);
      bless $mem->{$elt}, $type;
    }

    return($mem->{$elt});
  }
}

sub ar_copy
{
  my $type;
  my $object = {};
  my $array  = [];
  my $mem    = shift;

  foreach my $elt (@_)
  {
    $type = ref $elt;

    unless ($type)
    {
      push (@$array,$elt);
    }
    else
    {
      if ($type =~ /ARRAY/)
      {
        unless (defined($mem->{$elt}))
        {
	  $mem->{$elt} = [];
          @{$mem->{$elt}}=&ar_copy($mem,@$elt);
        }
      }
      elsif ($type =~ /HASH/)
      {
        unless (defined($mem->{$elt}))
        {
	  $mem->{$elt} = {};
          %{$mem->{$elt}}=&ha_copy($mem,$elt);
        }
      }
      elsif ($type =~ /SCALAR|REF/)
      {

        unless (defined($mem->{$elt}))
        {
	  $mem->{$elt}='';
          $mem->{$elt}=&sr_copy($mem,$$elt) unless (defined($mem->{$elt}));
        }

      }
      elsif ($type =~ /CODE/)
      {
        push (@$array,$elt);
      }
      else
      {
        unless (defined($mem->{$elt}))
        {
	  $mem->{$elt} = {};
          %{$mem->{$elt}} = &ha_copy($mem,$elt);
          bless $mem->{$elt}, $type;
        }
      }
      push (@$array,$mem->{$elt});
    }
  }

  return(@$array);
}

sub ha_copy
{
  my $type;
  my $array = {};
  my $mem   = shift;
  my $hash  = shift;

  foreach my $key (keys %$hash)
  {
    $type = ref $hash->{$key};

    unless ($type)
    {
      $array->{$key}=$hash->{$key};
    }
    else
    {
      if ($type =~ /ARRAY/)
      {
        unless (defined($mem->{$hash->{$key}}))
        {
	  $mem->{$hash->{$key}}= [];
          @{$mem->{$hash->{$key}}}=&ar_copy($mem,@{$hash->{$key}})
        }
      }
      elsif ($type =~ /HASH/)
      {
       unless (defined($mem->{$hash->{$key}}))
        {
	  $mem->{$hash->{$key}}= {};
	  %{$mem->{$hash->{$key}}}=&ha_copy($mem,$hash->{$key});
        }

      }
      elsif ($type =~ /SCALAR|REF/)
      {
        unless (defined($mem->{$hash->{$key}}))
	{
          $mem->{$hash->{$key}}='';
          $mem->{$hash->{$key}}=&sr_copy($mem,${$hash->{$key}});
        }
      }
      elsif ($type =~ /CODE/)
      {
        $mem->{$hash->{$key}}=$hash->{$key};
      }
      else
      {
	unless (defined($mem->{$hash->{$key}}))
        {
          $mem->{$hash->{$key}}={};
          %{$mem->{$hash->{$key}}}=&ha_copy($mem,$hash->{$key});
          bless $mem->{$hash->{$key}}, $type;
        }
      }
      $array->{$key}=$mem->{$hash->{$key}};
    }
  }

  return(%$array);
}

sub sr_copy
{
  my $mem  = shift;
  my $sr   = shift;
  my $object;

  return (\$sr) unless (ref $sr);
  $mem->{$sr} = &u_copy($sr,$mem)
  unless (defined($mem->{$sr}));

  $object = $mem->{$sr};

  return(\$object);
}

sub thorelli
{
  my $copy      = shift;
  my $intersect = shift;
  my $mem       = shift;
  my $thorelli  = shift;

  my $type   = ref $copy;

  no warnings 'uninitialized';
  $thorelli->{$copy}=1;

  if ($type  && $type !~ /SCALAR|CODE/)
  {
    if ($type =~ /REF/)
    {
      &thorelli($$copy,$intersect,$mem,$thorelli) unless ($thorelli->{$$copy});
    }
    elsif ($type =~ /ARRAY/)
    {
      foreach my $elt (@$copy)
      {
        &thorelli($elt,$intersect,$mem,$thorelli) unless ($thorelli->{$elt});
      }
    }
    else
    {
      foreach my $key (keys %$copy)
      {
        if (defined($intersect->{$key}))
        {
          $copy->{$mem->{$key}}=$copy->{$key};
          &thorelli($copy->{$mem->{$key}},$intersect,$mem,$thorelli) unless ($thorelli->{$copy->{$key}});
          delete $copy->{$key};
        }
        else
        {
          &thorelli($copy->{$key},$intersect,$mem,$thorelli) unless ($thorelli->{$copy->{$key}});
        }
      }
    }
  }
}

sub hexbytes
{
#  return(shift);
  unpack sprintf("H%d",2*length($_[0])),$_[0];
}

sub unhexbytes
{
#  return(shift);
  pack sprintf("H%d",length($_[0])),$_[0];
}

sub checktext
{
#  return(shift);
  my $text=shift;
  my $t=$text;

  $t =~ s/\n//gs;
  $t =~ s/'(A|H|P:\w+|R|S|V|L):\d+'//g;
  $t =~ s/'[0-9a-f]*'//g;
  $t =~ s/\[,*\]//g;
  $t =~ s/=>(,=>)+/=>/g;    # this line is added to avoid a segmentation fault under ETCH (debian distribution)
  $t =~ s/\{(=>,?)*\}//g;
  $t =~ s/=>(,=>)*//g;
  $t =~ s/\(,\s*(\{\})?\)\s*//s;

  if ($t)
  {
    warn "Datafile is corrupt: $text => $t\n";
    return(undef);
  }
  return($text);
}

'end Clone';

=head1 Clone.pm
=head1 SYNOPSIS

Clone copies entire PERL datastructures

=head1 DEPENDENCIES

None

=head1 DESCRIPTION

 Het package Clone kan een klasse van datastructuren ononderscheidbaar te copieren
 Dat wil zeggen: Er bestaat geen enkele dergelijke datastructuur waarvoor een procedure
 geschreven kan worden die in staat is het orgineel te onderscheiden van de Clone

 Voorwaarde is dat de data structuur slechts is samengesteld uit referenties naar hashes,
 referenties naar lists, en of text:

 BV:  {'aa' => ['a','b','c'], 'bb'=> 1, 'cc' => 2, 'dd' => ['d','e',{'f','g'}]}

=head1 METHODS

head2 clone()
over 2

  SYNOPSIS 

  $clone_instance2 = $clone_instance1->clone();

  DESCRIPTION

  hiermee kan iedere instantie van Clone zichzelf copieren

  NEEDS

  refkeys()   
  u_copy()    
  thorelli()

back 2

head2 copy()
over 2

  SYNOPSIS 

  $data2 = Clone::copy($data1);

  DESCRIPTION

  hiermee kan Clone een willekeurige datastructuur copieren

  NEEDS

  refkeys()   
  u_copy()    
  thorelli()

back 2

head2 copy2text()
over 2

  SYNOPSIS 

  $text = Clone::copy2text(data);

  DESCRIPTION


  hiermee kan Clone een willekeurige datastructuur omzetten naar
  textformaat (ascii) zonder dat de ooprspronkelijke datastructuur aangetast
  wordt.

  NEEDS

  linear() 
  outnumber()
  text()  

back 2

head2 convert2text()
over 2

  SYNOPSIS 

  $text = Clone::convert2text(data);

  DESCRIPTION

  Snellere versie van copy2text hiermee kan iedere instantie van Clone 
  een willekeurige datastructuur omzetten naar textformaat (ascii) 
  waarbij de oorspronkelijke datastructuur ernstig gecorumpeerd wordt.

  NEEDS

  linear ()
  outnumber()
  text()  

back 2

head2 fromtext()
over 2

  SYNOPSIS 

  $data = Clone::fromtext($text);

  DESCRIPTION

  hiermee kan Clone een door copy2text of convert2text
  geproduceerde text omzetten naar een echte Perl datastructuur.

  NEEDS

  hexbytes()     
  unhexbytes()   
  checktext()

back 2

head2 Private functies

over 2

  refkeys      : Loopt recursief door de datastructuur en maakt een lijst van referenties
  u_copy       : Maakt een oppervlakkige copie van de datastructuur, maakt gebruik van:
                   ha_copy (hash_copy), 
                   ar_copy (array_copy),
                   sr_copy (scalar/reference_copy). 
                 die elkaar recursief aanroepen.


  thorelli     : Zet de oppervlakkige copie om in een volledige copie.
  linear       : Kent aan iedere afzonderlijke waarde (token) in de datastructuur een
                 unieke identificatie toe (type) en produceert twee hash tabellen,
                 een van type-naar-token en een van token-naar-type.
  outnumber    : zet de type-naar-token-tabel om in een tabel die alle referenties naar een
                 partiele datastructuur gevuld met waarden omzet naar referenties naar
                 eenzelfde datastructuur maar dan gevuld met de corresponderende type aanduiding.

  text         : zet laatsgenoemde tabel om in text.
  hexbytes     : zet textbytes om naar hexadecimale notatie.
  unhexbytes   : zet hexadecimale notatie om naar textbytes.
  checktext    : checkt of de ingelezen text veilig door de PERL eval functie geevalueerd kan worden.

back 2

=head1 BUGS

Geen bekende bugs (mogelijk wel onbekende)

=head1 AUTHOR

Rob Vossen <r.n.f.vossen@uvt.nl>

=head1 COPYRIGHT AND LICENCE

Copyright (C) 2004 by Rob Vossen

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.4 or,
at your option, any later version of Perl 5 you may have available.

=cut
